"""Tests for common functions and classes for I/O."""

# Copyright 2014, 2015 Matt Shannon

# This file is part of htk_io.
# See `License` for details of license and warranty.

import unittest
import doctest

import htk_io.base

def load_tests(loader, tests, ignore):
    tests.addTests(doctest.DocTestSuite(htk_io.base))
    return tests

# FIXME : add tests
class BaseTest(unittest.TestCase):
    pass

if __name__ == '__main__':
    unittest.main()



"""Functions for reading and writing HTK / HTS alignment files.

HTK / HTS alignment files are often referred to as "label" files.

In this module an alignment is a sequence of segments, where each segment is a
tuple

    (startTime, endTime, label, subAlignment)

where `startTime` and `endTime` are the integer-valued numbers of frames,
`label` is an arbitrary label associated with the segment and `subAlignment` is
an alignment or None (to specify no sub-alignment).
"""

# Copyright 2014, 2015 Matt Shannon

# This file is part of htk_io.
# See `License` for details of license and warranty.

import os

from htk_io.base import LineIo
from htk_io.base import DirReader

class SimpleAlignmentIo(LineIo):
    """Reads and writes 1-level HTK-style alignment files.

    The amount of time between one frame and the next is specified by the float
    `framePeriod` in seconds.

    In most cases it is probably preferable to use `AlignmentIo` instead of
    this class since it supports multilevel alignments.
    """
    def __init__(self, framePeriod):
        self.framePeriod = framePeriod

        if self.framePeriod < 1e-7:
            # N.B. I believe write-then-read is guaranteed to recover original
            #   integer-valued segment timings as long as framePeriod >= 1e-7
            raise RuntimeError(
                'writing alignment files may be lossy since specified'
                ' frame period is extremely small'
            )

    def writeLines(self, alignment):
        """Writes the lines of a 1-level HTK-style alignment file.

        Use `writeFile` method to write an actual file.

        See the 1-level alignment examples in the documentation for
        `AlignmentIo.writeLines` for examples of usage of this method.
        """
        divisor = self.framePeriod * 1e7

        alignmentLines = []
        for startTime, endTime, label, subAlignment in alignment:
            assert subAlignment is None
            startTicks = int(round(startTime * divisor))
            endTicks = int(round(endTime * divisor))
            alignmentLines.append('%s %s %s' % (startTicks, endTicks, label))

        return alignmentLines

    def readLines(self, alignmentLines):
        """Reads the lines of a 1-level HTK-style alignment file.

        Use `readFile` method to read an actual file.

        See the 1-level alignment examples in the documentation for
        `AlignmentIo.readLines` for examples of usage of this method.
        """
        divisor = self.framePeriod * 1e7

        alignment = []
        for line in alignmentLines:
            startTicks, endTicks, label = line.strip().split(None, 2)
            startTime = int(round(int(startTicks) / divisor))
            endTime = int(round(int(endTicks) / divisor))
            alignment.append((startTime, endTime, label, None))

        return alignment

def flatten(alignment, checkRecover=True):
    """Converts a hierarchical alignment to a flat alignment.

    The flat alignment is a 1-level alignment where each label is a tuple.
    Each element of the tuple corresponds to a different level of the
    hierarchical alignment, with the 0th element of the tuple corresponding to
    the innermost (i.e. the most frequently changing) level of the hierarchical
    alignment.

    For example for a 2-level alignment:

    >>> import htk_io.alignment as alio
    >>> alio.flatten([
    ...     (0, 2, 'a', [
    ...         (0, 1, 'X', None),
    ...         (1, 2, 'Y', None),
    ...     ]),
    ...     (2, 3, 'b', [
    ...         (2, 3, 'Z', None),
    ...     ]),
    ... ]) == [
    ...     (0, 1, ('X', 'a'), None),
    ...     (1, 2, ('Y',), None),
    ...     (2, 3, ('Z', 'b'), None),
    ... ]
    True

    Any sub-alignments should either be non-empty lists or None:

    >>> alio.flatten([
    ...     (0, 1, 'a', []),
    ... ])
    Traceback (most recent call last):
        ...
    RuntimeError: sub-alignment was neither a non-empty list nor None: []
    >>> alio.flatten([
    ...     (0, 1, 'a', [
    ...         (0, 1, 'A', []),
    ...     ]),
    ... ])
    Traceback (most recent call last):
        ...
    RuntimeError: sub-alignment was neither a non-empty list nor None: []

    However for convenience the main alignment is allowed to be an empty list:
    >>> alio.flatten([])
    []

    The hierarchical alignment should be of a consistent depth.
    If `checkRecover` is True then a check is performed to verify that the
    original hierarchical alignment is recoverable from the flat alignment.
    This is the case whenever the restrictions mentioned above are adhered to.

    >>> alio.flatten([
    ...     (0, 2, 'a', [
    ...         (0, 1, 'A', None),
    ...         (1, 2, 'B', None),
    ...     ]),
    ...     (2, 3, 'b', None),
    ... ]) # doctest: +ELLIPSIS
    Traceback (most recent call last):
        ...
    RuntimeError: alignment could not be recovered...
    """
    flatAlignment = []
    for startTime, endTime, label, subAlignment in alignment:
        if not subAlignment:
            if subAlignment is not None:
                raise RuntimeError(
                    'sub-alignment was neither a non-empty list nor None: %s' %
                    subAlignment
                )
            flatAlignmentSub = [(startTime, endTime, (), subAlignment)]
        else:
            flatAlignmentSub = flatten(
                subAlignment,
                checkRecover=False
            )
        assert flatAlignmentSub
        for (
            entryIndex, (startTimeSub, endTimeSub, labelTuple, subAlignmentSub)
        ) in enumerate(flatAlignmentSub):
            flatAlignment.append((
                startTimeSub,
                endTimeSub,
                labelTuple if entryIndex > 0 else labelTuple + (label,),
                subAlignmentSub
            ))

    if checkRecover:
        if unflatten(flatAlignment) != alignment:
            raise RuntimeError(
                'alignment could not be recovered from flattened alignment'
                ' (check alignment has consistent depth)'
            )

    return flatAlignment

def unflatten(flatAlignment):
    """Converts a flat alignment to a hierarchical alignment.

    This function is the inverse of `flatten`.

    For example for a 2-level alignment:

    >>> import htk_io.alignment as alio
    >>> alio.unflatten([
    ...     (0, 1, ('X', 'a'), None),
    ...     (1, 2, ('Y',), None),
    ...     (2, 3, ('Z', 'b'), None),
    ... ]) == [
    ...     (0, 2, 'a', [
    ...         (0, 1, 'X', None),
    ...         (1, 2, 'Y', None),
    ...     ]),
    ...     (2, 3, 'b', [
    ...         (2, 3, 'Z', None),
    ...     ]),
    ... ]
    True
    """
    if not flatAlignment:
        assert flatAlignment is not None
        alignment = []
    else:
        numLevels = len(flatAlignment[0][2])
        assert numLevels >= 1

        currLabels = [ None for _ in range(numLevels) ]
        currSubAlignments = [ [] for _ in range(numLevels) ]
        for entry, entryNext in zip(flatAlignment, flatAlignment[1:] + [None]):
            currNumLevels = len(entry[2])
            assert 1 <= currNumLevels <= numLevels

            startTime, endTime, labelTuple, subAlignment = entry
            assert subAlignment is None

            currLabels[:currNumLevels] = labelTuple

            numFreeze = numLevels if entryNext is None else len(entryNext[2])
            for freezeIndex in range(numFreeze):
                label = currLabels[freezeIndex]
                if freezeIndex >= 1:
                    subAlignmentSeg = currSubAlignments[freezeIndex - 1]
                    startTimeSeg = subAlignmentSeg[0][0]
                    endTimeSeg = subAlignmentSeg[-1][1]
                    currSubAlignments[freezeIndex].append(
                        (startTimeSeg, endTimeSeg, label, subAlignmentSeg)
                    )
                    currLabels[freezeIndex] = None
                    currSubAlignments[freezeIndex - 1] = []
                else:
                    currSubAlignments[freezeIndex].append(
                        (startTime, endTime, label, None)
                    )
                    currLabels[freezeIndex] = None

        alignment = currSubAlignments[numLevels - 1]

    return alignment

class AlignmentIo(LineIo):
    """Reads and writes HTK-style alignment files.

    Example usage:

    >>> import htk_io.alignment as alio
    >>> alignmentIo = alio.AlignmentIo(framePeriod=1.0)
    >>> alignmentIo.writeLines([
    ...     (0, 1, 'the', None),
    ...     (1, 2, 'cat', None),
    ...     (2, 5, 'cat', None),
    ...     (5, 6, 'sat', None),
    ... ]) == [
    ...     '0 10000000 the',
    ...     '10000000 20000000 cat',
    ...     '20000000 50000000 cat',
    ...     '50000000 60000000 sat',
    ... ]
    True
    >>> alignmentIo.readLines([
    ...     '0 10000000 the',
    ...     '10000000 20000000 cat',
    ...     '20000000 50000000 cat',
    ...     '50000000 60000000 sat',
    ... ]) == [
    ...     (0, 1, 'the', None),
    ...     (1, 2, 'cat', None),
    ...     (2, 5, 'cat', None),
    ...     (5, 6, 'sat', None),
    ... ]
    True

    The amount of time between one frame and the next is specified by the float
    `framePeriod` in seconds.

    >>> alignmentIo2 = alio.AlignmentIo(framePeriod=0.5)
    >>> alignmentIo2.writeLines([
    ...     (0, 2, 'the', None),
    ...     (2, 4, 'cat', None),
    ...     (4, 10, 'cat', None),
    ...     (10, 12, 'sat', None),
    ... ]) == [
    ...     '0 10000000 the',
    ...     '10000000 20000000 cat',
    ...     '20000000 50000000 cat',
    ...     '50000000 60000000 sat',
    ... ]
    True
    >>> alignmentIo2.readLines([
    ...     '0 10000000 the',
    ...     '10000000 20000000 cat',
    ...     '20000000 50000000 cat',
    ...     '50000000 60000000 sat',
    ... ]) == [
    ...     (0, 2, 'the', None),
    ...     (2, 4, 'cat', None),
    ...     (4, 10, 'cat', None),
    ...     (10, 12, 'sat', None),
    ... ]
    True

    A check is performed to ensure that no information will be lost by the
    conversion from number of frames to units of 1e-7 seconds that happens
    during alignment writing:

    >>> alignmentIo3 = alio.AlignmentIo(framePeriod=5e-8)
    ... # doctest: +ELLIPSIS
    Traceback (most recent call last):
        ...
    RuntimeError: writing alignment files may be lossy...

    Alignments not starting at time zero are supported:

    >>> alignmentIo.writeLines([
    ...     (5, 6, 'the', None),
    ...     (6, 8, 'cat', None),
    ... ]) == [
    ...     '50000000 60000000 the',
    ...     '60000000 80000000 cat',
    ... ]
    True
    >>> alignmentIo.readLines([
    ...     '50000000 60000000 the',
    ...     '60000000 80000000 cat',
    ... ]) == [
    ...     (5, 6, 'the', None),
    ...     (6, 8, 'cat', None),
    ... ]
    True
    >>> alignmentIo.writeLines([
    ...     (-6, -5, 'the', None),
    ...     (-5, 4, 'cat', None),
    ... ]) == [
    ...     '-60000000 -50000000 the',
    ...     '-50000000 40000000 cat',
    ... ]
    True
    >>> alignmentIo.readLines([
    ...     '-60000000 -50000000 the',
    ...     '-50000000 40000000 cat',
    ... ]) == [
    ...     (-6, -5, 'the', None),
    ...     (-5, 4, 'cat', None),
    ... ]
    True

    It supports multilevel alignments.
    For example for a 2-level alignment:

    >>> alignmentIo.writeLines([
    ...     (0, 3, 'the', [
    ...         (0, 2, 'X', None),
    ...         (2, 3, 'Y', None),
    ...     ]),
    ...     (3, 6, 'cat', [
    ...         (3, 4, 'Y', None),
    ...         (4, 6, 'X', None),
    ...     ]),
    ... ]) == [
    ...     '0 20000000 X the',
    ...     '20000000 30000000 Y',
    ...     '30000000 40000000 Y cat',
    ...     '40000000 60000000 X',
    ... ]
    True
    >>> alignmentIo.readLines([
    ...     '0 20000000 X the',
    ...     '20000000 30000000 Y',
    ...     '30000000 40000000 Y cat',
    ...     '40000000 60000000 X',
    ... ]) == [
    ...     (0, 3, 'the', [
    ...         (0, 2, 'X', None),
    ...         (2, 3, 'Y', None),
    ...     ]),
    ...     (3, 6, 'cat', [
    ...         (3, 4, 'Y', None),
    ...         (4, 6, 'X', None),
    ...     ]),
    ... ]
    True

    It works even for long utterances:

    >>> alignmentIo.writeLines([
    ...     (0, 360000, 'a', None),
    ... ]) == [
    ...     '0 3600000000000 a',
    ... ]
    True
    >>> alignmentIo.readLines([
    ...     '0 3600000000000 a',
    ... ]) == [
    ...     (0, 360000, 'a', None),
    ... ]
    True
    >>> alignmentIo.writeLines([
    ...     (-360000, 0, 'a', None),
    ... ]) == [
    ...     '-3600000000000 0 a',
    ... ]
    True
    >>> alignmentIo.readLines([
    ...     '-3600000000000 0 a',
    ... ]) == [
    ...     (-360000, 0, 'a', None),
    ... ]
    True
    """
    def __init__(self, framePeriod, levelSep=None):
        self.framePeriod = framePeriod
        self.levelSep = levelSep

        self.simpleIo = SimpleAlignmentIo(self.framePeriod)

    def writeLines(self, alignment):
        """Writes the lines of an HTK-style alignment file.

        Use `writeFile` method to write an actual file.
        """
        levelSep = ' ' if self.levelSep is None else self.levelSep

        flatAlignment = flatten(alignment)
        rawAlignment = [
            (startTime, endTime, levelSep.join(labelTuple), subAlignment)
            for startTime, endTime, labelTuple, subAlignment in flatAlignment
        ]
        alignmentLines = self.simpleIo.writeLines(rawAlignment)
        return alignmentLines

    def readLines(self, alignmentLines):
        """Reads the lines of an HTK-style alignment file.

        Use `readFile` method to read an actual file.
        """
        rawAlignment = self.simpleIo.readLines(alignmentLines)
        flatAlignment = [
            (startTime, endTime, label.split(self.levelSep), subAlignment)
            for startTime, endTime, label, subAlignment in rawAlignment
        ]
        alignment = unflatten(flatAlignment)
        return alignment

def mapAlignmentLabels(alignment, labelMaps):
    """Transforms an alignment by mapping the labels at each depth.

    Labels at the highest level are mapped using labelMaps[0], and labels at
    the second highest level are mapped using labelMaps[1], etc.
    """
    if not labelMaps:
        return alignment
    else:
        labelMap = labelMaps[0]
        subLabelMaps = labelMaps[1:]
        return [
            (
                startTime,
                endTime,
                labelMap(label),
                mapAlignmentLabels(subAlignment, subLabelMaps)
            )
            for startTime, endTime, label, subAlignment in alignment
        ]

# (FIXME : support self.inv.inv? How to encode general invertible transforms?)
class AlignmentLabelTransform(object):
    """Transforms an alignment by mapping the labels at each depth.

    Labels at the highest level are mapped using labelMaps[0], and labels at
    the second highest level are mapped using labelMaps[1], etc.

    Example usage:

    >>> import htk_io.alignment as alio
    >>> alignmentTransform = alio.AlignmentLabelTransform(
    ...     [lambda label: label == 'cat']
    ... )
    >>> alignmentTransform([
    ...     (0, 1, 'the', None),
    ...     (1, 2, 'cat', None),
    ...     (2, 5, 'cat', None),
    ...     (5, 6, 'sat', None),
    ... ]) == [
    ...     (0, 1, False, None),
    ...     (1, 2, True, None),
    ...     (2, 5, True, None),
    ...     (5, 6, False, None),
    ... ]
    True

    If each of the label transforms has an `inv` method to compute the inverse
    then so does the overall alignment transform:

    >>> class SimpleTransform(object):
    ...     def __call__(self, string):
    ...         return list(string)
    ...
    ...     def inv(self, _list):
    ...         return ''.join(_list)
    ...
    >>> alignmentTransform = alio.AlignmentLabelTransform([SimpleTransform()])
    >>> alignmentTransform([
    ...     (0, 1, 'the', None),
    ...     (1, 2, 'cat', None),
    ...     (2, 5, 'cat', None),
    ...     (5, 6, 'sat', None),
    ... ]) == [
    ...     (0, 1, ['t', 'h', 'e'], None),
    ...     (1, 2, ['c', 'a', 't'], None),
    ...     (2, 5, ['c', 'a', 't'], None),
    ...     (5, 6, ['s', 'a', 't'], None),
    ... ]
    True
    >>> alignmentTransform.inv([
    ...     (0, 1, ['t', 'h', 'e'], None),
    ...     (1, 2, ['c', 'a', 't'], None),
    ...     (2, 5, ['c', 'a', 't'], None),
    ...     (5, 6, ['s', 'a', 't'], None),
    ... ]) == [
    ...     (0, 1, 'the', None),
    ...     (1, 2, 'cat', None),
    ...     (2, 5, 'cat', None),
    ...     (5, 6, 'sat', None),
    ... ]
    True
    """
    def __init__(self, labelTransforms):
        self.labelTransforms = labelTransforms

        try:
            self.labelTransformInvs = [
                labelTransform.inv
                for labelTransform in self.labelTransforms
            ]
        except:
            pass

    def __call__(self, alignment):
        return mapAlignmentLabels(alignment, self.labelTransforms)

    def inv(self, alignmentNew):
        return mapAlignmentLabels(alignmentNew, self.labelTransformInvs)



"""Functions for reading and writing HTK / HTS question files."""

# Copyright 2014, 2015 Matt Shannon

# This file is part of htk_io.
# See `License` for details of license and warranty.

import re
import fnmatch

from htk_io.misc import stripQuotes, addQuotes, verifiedRead

def getQuesRe(quesPats):
    """Converts a list of question patterns to a regular expression."""
    # N.B. -7 is to get rid of superfluous '\Z(?ms)' at end of string
    quesReStrs = [ fnmatch.translate(quesPat)[:-7] for quesPat in quesPats ]
    quesRe = re.compile(('|'.join(quesReStrs))+'$')
    return quesRe

def getQuesReDict(questions):
    """Returns a dictionary mapping a question id to a regular expression."""
    quesReDict = dict()
    for quesId, quesPats in questions:
        assert quesId not in quesReDict
        quesReDict[quesId] = getQuesRe(quesPats)

    return quesReDict

def parseQuestionLines(lines, isTreeFile=False):
    questions = []
    restIndex = len(lines)
    for lineIndex, line in enumerate(lines):
        line = line.strip()
        fields = line.split()
        if len(fields) > 0:
            if fields[0] == 'QS':
                if isTreeFile:
                    assert len(fields) == 5
                    assert fields[2] == '{'
                    assert fields[4] == '}'
                    quesId = fields[1]
                    quesPats = map(stripQuotes, fields[3].split(','))
                    questions.append((quesId, quesPats))
                else:
                    assert len(fields) == 3
                    assert fields[2][0] == '{'
                    assert fields[2][-1] == '}'
                    quesId = stripQuotes(fields[1])
                    quesPats = fields[2][1:-1].split(',')
                    questions.append((quesId, quesPats))
            else:
                restIndex = lineIndex
                break

    return questions, lines[restIndex:]

def readQuestionLines(lines, isTreeFile=False):
    questions, linesRest = parseQuestionLines(lines, isTreeFile=isTreeFile)
    assert not linesRest
    return questions

def writeQuestionLines(questions, isTreeFile=False):
    outLines = []
    for quesId, quesPats in questions:
        if isTreeFile:
            quesPatsStr = ','.join(map(addQuotes, quesPats))
            outLines.append('QS %s { %s }' % (quesId, quesPatsStr))
        else:
            quesPatsStr = ','.join(quesPats)
            outLines.append('QS %s {%s}' % (addQuotes(quesId), quesPatsStr))

    return outLines

def readQuesFile(quesFile):
    """Reads a question file."""
    lines = [ line.rstrip('\n') for line in open(quesFile, 'U') ]
    questions = readQuestionLines(lines)
    return questions

def readQuesFileVerifying(quesFile):
    """Reads a question file and verifies that it was read correctly.

    Verifies that reconstructing the question file from the parsed output
    reproduces the original question file up to certain whitespace
    substitutions.
    """
    lines = [ line.rstrip('\n') for line in open(quesFile, 'U') ]
    questions = verifiedRead(readQuestionLines, writeQuestionLines, lines)
    return questions

def writeQuesFile(questions, quesFile):
    """Writes a question file."""
    lines = writeQuestionLines(questions)
    with open(quesFile, 'w') as f:
        for line in lines:
            f.write(line)
            f.write('\n')



"""Common functions and classes for I/O."""

# Copyright 2014, 2015 Matt Shannon

# This file is part of htk_io.
# See `License` for details of license and warranty.

import os

class Io(object):
    """An abstract class for reading and writing files."""
    pass

class LineIo(Io):
    """An abstract class for reading and writing line-based files."""
    def writeFile(self, filename, obj):
        lines = self.writeLines(obj)
        with open(filename, 'w') as f:
            for line in lines:
                f.write(line)
                f.write('\n')

    def readFile(self, filename):
        lines = [ line.rstrip('\n') for line in open(filename, 'U') ]
        return self.readLines(lines)

# (FIXME : add corresponding write method and rename class
#    (after deciding how to encode general invertible transforms nicely))
class DirReader(object):
    """This class reads files with a common extension from a directory.

    An instance of this class can be used as a function from file base name to
    object.
    For a given base name, the object returned is read using `io` from the file
    in directory `readDir` with extension `ext`.

    Example usage (assumes the current directory contains the "example"
    subdirectory included in source code for this package):

    >>> from htk_io.base import DirReader
    >>> import htk_io.alignment as alio
    >>> alignmentIo = alio.AlignmentIo(framePeriod=0.005)
    >>> getAlignment = DirReader(alignmentIo, 'example', 'lab')
    >>> getAlignment('simple') == [
    ...     (0, 10, 'apple', None),
    ...     (10, 41, 'pears', None),
    ... ]
    True

    For a 2-level alignment:

    >>> getAlignment('simple-2-level') == [
    ...     (0, 10, 'apple', [
    ...         (0, 2, 'a', None),
    ...         (2, 3, 'p', None),
    ...         (3, 5, 'p', None),
    ...         (5, 8, 'l', None),
    ...         (8, 10, 'e', None),
    ...     ]),
    ...     (10, 41, 'pears', [
    ...         (10, 11, 'p', None),
    ...         (11, 13, 'e', None),
    ...         (13, 14, 'a', None),
    ...         (14, 27, 'r', None),
    ...         (27, 41, 's', None),
    ...     ]),
    ... ]
    True

    Or even a 3-level alignment:

    >>> alignmentIo2 = alio.AlignmentIo(framePeriod=1.0)
    >>> getAlignment2 = DirReader(alignmentIo2, 'example', 'lab')
    >>> getAlignment2('example-3-level') == [
    ...     (0, 8, '0', [
    ...         (0, 3, 'a', [
    ...             (0, 1, 'A', None),
    ...             (1, 2, 'B', None),
    ...             (2, 3, 'C', None),
    ...         ]),
    ...         (3, 5, 'b', [
    ...             (3, 4, 'D', None),
    ...             (4, 5, 'E', None),
    ...         ]),
    ...         (5, 6, 'c', [
    ...             (5, 6, 'F', None),
    ...         ]),
    ...         (6, 8, 'c', [
    ...             (6, 7, 'G', None),
    ...             (7, 8, 'H', None),
    ...         ]),
    ...     ]),
    ...     (8, 10, '1', [
    ...         (8, 10, 'd', [
    ...             (8, 10, 'I', None),
    ...         ]),
    ...     ]),
    ... ]
    True
    """
    def __init__(self, io, readDir, ext, transform=None):
        self.io = io
        self.readDir = readDir
        self.ext = ext
        self.transform = transform

    def __call__(self, base):
        objFile = os.path.join(self.readDir, '%s.%s' % (base, self.ext))
        obj = self.io.readFile(objFile)
        if self.transform is not None:
            obj = self.transform(obj)
        return obj



"""Helper functions for I/O."""

# Copyright 2014, 2015 Matt Shannon

# This file is part of htk_io.
# See `License` for details of license and warranty.

import re

def stripQuotes(s):
    assert len(s) >= 2 and s[0] == '"' and s[-1] == '"'

    return s[1:-1]

def addQuotes(s):
    return '"%s"' % s

whitespaceRe = re.compile(r'\s+')
whitespaceEolRe = re.compile(r'\s+$')

def normalizeWhitespace(lines):
    outLines = []
    for line in lines:
        outLine = line
        outLine = re.sub(whitespaceRe, ' ', outLine)
        outLine = re.sub(whitespaceEolRe, '', outLine)
        if outLine != '':
            outLines.append(outLine)
    return outLines

def verifiedRead(read, write, lines, unpack=False):
    data = read(lines)
    if unpack:
        linesOut = write(*data)
    else:
        linesOut = write(data)

    lines = normalizeWhitespace(lines)
    linesOut = normalizeWhitespace(linesOut)

    if linesOut != lines:
        for line, lineOut in zip(lines, linesOut):
            if line != lineOut:
                print line
                print ' !='
                print lineOut
                raise RuntimeError('verified read failed')
        assert len(line) != len(lineOut)
        print 'lengths differ: %s vs %s' % (len(line), len(lineOut))
        raise RuntimeError('verified read failed')

    return data



"""Functions for reading and writing raw vector sequence files."""

# Copyright 2014, 2015 Matt Shannon

# This file is part of htk_io.
# See `License` for details of license and warranty.

import numpy as np

# FIXME : add tests

class VecSeqIo(object):
    """Reads and writes raw vector sequence files.

    This raw format is used by HTS for speech parameter files (as well as by
    the Speech Processing Toolkit (SPTK) for lots of purposes).
    """
    def __init__(self, vecSize, dtypeFile=np.float32):
        self.vecSize = vecSize
        self.dtypeFile = dtypeFile

    def readFile(self, vecSeqFile):
        """Reads a raw vector sequence file.

        The dtype of the returned numpy array is always the numpy default
        np.float, which may be 32-bit or 64-bit depending on architecture, etc.
        """
        return np.reshape(
            np.fromfile(vecSeqFile, dtype=self.dtypeFile),
            (-1, self.vecSize)
        ).astype(np.float)

    def writeFile(self, vecSeqFile, vecSeq):
        """Writes a raw vector sequence file."""
        vecSeq.astype(self.dtypeFile).tofile(vecSeqFile)

class VecSeqToTraj(object):
    """Extracts one particular trajectory from a vector sequence."""
    def __init__(self, vecIndex):
        self.vecIndex = vecIndex

    def __call__(self, vecSeq):
        # (copy here allows larger array to be garbage collected)
        return vecSeq[:, self.vecIndex].copy()



#!/usr/bin/python
"""Fixes a mapping from leaf macro id to leaf index for an HTK / HTS tree file.

For example, might map "mgc_s2_23" to "0".
"""

# Copyright 2014, 2015 Matt Shannon

# This file is part of htk_io.
# See `License` for details of license and warranty.

import sys
import argparse

import htk_io.tree as tio

def main(argv):
    parser = argparse.ArgumentParser(description=__doc__.split('\n')[0])
    parser.add_argument(
        dest='treeFile', metavar='TREE',
        help='an HTK / HTS tree file (e.g. "mgc.inf")'
    )
    args = parser.parse_args(argv[1:])

    questions, streamSpecedTrees = tio.readTreeFileVerifying(args.treeFile)

    leafMacroIdSet = set()
    for streamSpec, tree in streamSpecedTrees:
        for leaf in tree.leaves:
            leafMacroIdSet.add(leaf.macroId)

    for leafIndex, leafMacroId in enumerate(sorted(leafMacroIdSet)):
        print '%s %s' % (leafMacroId, leafIndex)

if __name__ == '__main__':
    main(sys.argv)



#!/usr/bin/python
"""Maps each label in each alignment file using a user-specified mapping.

If an input alignment file has multiple levels then only the last (i.e. the
innermost, i.e the fastest-changing) level is considered.
"""

# Copyright 2014, 2015 Matt Shannon

# This file is part of htk_io.
# See `License` for details of license and warranty.

import os
import sys
import argparse

import htk_io.alignment as alio

def readLabelMapFile(labelMapFile):
    labelMapDict = dict()
    for line in open(labelMapFile):
        labelKey, labelValue = line.rstrip('\n').split()
        if labelKey in labelMapDict:
            raise RuntimeError('multiple values given for key %s' % labelKey)
        labelMapDict[labelKey] = labelValue

    return labelMapDict

def mapAlignment(labelMapDict, alignment):
    alignmentNew = []
    for startTime, endTime, label, _ in alignment:
        labelNew = labelMapDict[label]
        alignmentNew.append((startTime, endTime, labelNew, None))

    return alignmentNew

def main(argv):
    parser = argparse.ArgumentParser(description=__doc__.split('\n')[0])
    parser.add_argument(
        '--alignment_suffix', dest='alignmentSuffix',
        metavar='ALIGNSUFFIX',
        default='lab',
        help='suffix for alignment files'
    )
    parser.add_argument(
        dest='labelMapFile', metavar='LABELMAP',
        help=('file specifying label mapping'
              ' (two columns separated by whitespace)')
    )
    parser.add_argument(
        dest='alignmentDirIn', metavar='ALIGNDIRIN',
        help='directory to read input alignments from'
    )
    parser.add_argument(
        dest='uttIdsFile', metavar='UTTIDSFILE',
        help=('file containing a list of utterance ids'
              ' (e.g. one line might be "cmu_us_arctic_slt_a0001")')
    )
    parser.add_argument(
        dest='alignmentDirOut', metavar='ALIGNDIROUT',
        help='directory to write output alignments to'
    )
    args = parser.parse_args(argv[1:])

    uttIds = [ line.strip() for line in open(args.uttIdsFile) ]

    alignmentIo = alio.AlignmentIo(framePeriod=1e-7)

    labelMapDict = readLabelMapFile(args.labelMapFile)
    print '(read label map with %s entries)' % len(labelMapDict)

    print '(writing output to directory %s)' % args.alignmentDirOut
    for uttId in uttIds:
        alignmentFileIn = os.path.join(
            args.alignmentDirIn,
            '%s.%s' % (uttId, args.alignmentSuffix)
        )
        alignmentFileOut = os.path.join(
            args.alignmentDirOut,
            '%s.%s' % (uttId, args.alignmentSuffix)
        )

        alignment = alignmentIo.readFile(alignmentFileIn)
        alignmentNew = mapAlignment(labelMapDict, alignment)
        alignmentIo.writeFile(alignmentFileOut, alignmentNew)

if __name__ == '__main__':
    main(sys.argv)



#!/usr/bin/python
"""Maps each (label, sublabel) in each alignment file to a tree leaf macro id.

Given an HTS demo-style decision tree file and a collection of two-level
(label, sublabel) alignment files, this command maps each (label, sublabel)
pair to the corresponding leaf macro id as specified by the decision tree.
The result is a collection of 1-level alignment files where the labels are leaf
macro ids.

Two-level (label, sublabel) alignment files suitable for input to this command
may be obtained using HTS's HSMMAlign command with the -f flag.
"""

# Copyright 2014, 2015 Matt Shannon

# This file is part of htk_io.
# See `License` for details of license and warranty.

import os
import sys
import argparse

import htk_io.alignment as alio
import htk_io.ques as qio
import htk_io.tree as tio

def mapAlignment(alignment, navTrees, subLabelStrEnds):
    alignmentNew = []
    for startTime, endTime, label, subAlignment in alignment:
        assert len(subAlignment) == len(subLabelStrEnds)

        for subLabelIndex, (subStartTime, subEndTime, subLabelStr, _) in (
            enumerate(subAlignment)
        ):
            assert subLabelStr.endswith(subLabelStrEnds[subLabelIndex])

            leaf = navTrees[subLabelIndex].getLeaf(label)
            alignmentNew.append((subStartTime, subEndTime, leaf.macroId, None))

    return alignmentNew

def main(argv):
    parser = argparse.ArgumentParser(description=__doc__.split('\n')[0])
    parser.add_argument(
        '--alignment_suffix', dest='alignmentSuffix',
        metavar='ALIGNSUFFIX',
        default='lab',
        help='suffix for alignment files (e.g. "lab")'
    )
    parser.add_argument(
        '--sublabel_pat', dest='subLabelStrEndPat',
        metavar='PAT',
        default='[%d]',
        help=('printf-style pattern which, when expanded by replacing %%d with'
              ' a sublabel index >= 2, specifies the last part of the sublabel'
              ' string used in the input alignment files'
              ' (e.g. "[%%d]")')
    )
    parser.add_argument(
        '--stream_pat_pat', dest='streamSpecPat',
        metavar='PATPAT',
        default='{*}[%d].stream[1]',
        help=('printf-style pattern which, when expanded by replacing %%d with'
              ' a sublabel index >= 2, specifies the HTK macro pattern'
              ' identifying the desired tree for a given sublabel'
              ' (e.g. for HTS demo a value of "{*}[%%d].stream[1]" would use'
              ' trees from stream 1, which is the mgc stream)')
    )
    parser.add_argument(
        '--num_sublabels', dest='numSubLabels', metavar='NUMSUBLABELS',
        default=5, type=int,
        help=('number of sublabels (also sometimes known as "states") in input'
              ' alignments (e.g. "5")')
    )
    parser.add_argument(
        dest='treeFile', metavar='TREE',
        help='HTS demo-style decision tree file (e.g. "mgc.inf")'
    )
    parser.add_argument(
        dest='alignmentDirIn', metavar='ALIGNDIRIN',
        help='directory to read input alignments from'
    )
    parser.add_argument(
        dest='uttIdsFile', metavar='UTTIDSFILE',
        help=('file containing a list of utterance ids'
              ' (e.g. one line might be "cmu_us_arctic_slt_a0001")')
    )
    parser.add_argument(
        dest='alignmentDirOut', metavar='ALIGNDIROUT',
        help='directory to write output alignments to'
    )
    args = parser.parse_args(argv[1:])

    uttIds = [ line.strip() for line in open(args.uttIdsFile) ]

    subLabelStrEnds = [
        args.subLabelStrEndPat % (subLabelIndex + 2)
        for subLabelIndex in range(args.numSubLabels)
    ]

    alignmentIo = alio.AlignmentIo(framePeriod=1e-7)

    def subLabelIndexToStreamSpec(subLabelIndex):
        return args.streamSpecPat % (subLabelIndex + 2)

    questions, streamSpecedTrees = tio.readTreeFileVerifying(args.treeFile)
    quesReDict = qio.getQuesReDict(questions)

    navTrees = []
    for subLabelIndex in range(args.numSubLabels):
        streamSpecDesired = subLabelIndexToStreamSpec(subLabelIndex)
        foundTrees = [
            tree
            for streamSpec, tree in streamSpecedTrees
            if streamSpec == streamSpecDesired
        ]
        assert len(foundTrees) == 1
        navTree = tio.NavBinaryTree(quesReDict, foundTrees[0])
        navTrees.append(navTree)

    numLeaves = sum([ len(navTree.tree.leaves) for navTree in navTrees ])
    print '(found %s leaves)' % numLeaves

    print '(writing output to directory %s)' % args.alignmentDirOut
    for uttId in uttIds:
        alignmentFileIn = os.path.join(
            args.alignmentDirIn,
            '%s.%s' % (uttId, args.alignmentSuffix)
        )
        alignmentFileOut = os.path.join(
            args.alignmentDirOut,
            '%s.%s' % (uttId, args.alignmentSuffix)
        )

        alignment = alignmentIo.readFile(alignmentFileIn)
        alignmentNew = mapAlignment(alignment, navTrees, subLabelStrEnds)
        alignmentIo.writeFile(alignmentFileOut, alignmentNew)

if __name__ == '__main__':
    main(sys.argv)



"""Tests for functions for reading and writing HTK / HTS alignment files."""

# Copyright 2014, 2015 Matt Shannon

# This file is part of htk_io.
# See `License` for details of license and warranty.

import unittest
import doctest
import random
from numpy.random import randint, randn

import htk_io.alignment as alio

def load_tests(loader, tests, ignore):
    tests.addTests(doctest.DocTestSuite(alio))
    return tests

def gen_label(size=None, minSize=1, alphabet=None):
    if size is None:
        size = minSize + randint(2)
    if alphabet is None:
        alphabet = ['a', 'b']
    label = ''.join([
        random.choice(alphabet)
        for _ in range(size)
    ])
    return label

def gen_alignment(startTimeInit=None, numLevels=1,
                  size=None, minSize=0, minSizeSub=1, minDur=0,
                  labelMinSize=1, labelAlphabet=None):
    """Returns an alignment."""
    if startTimeInit is None:
        startTimeInit = random.choice([0, randint(-10, 11)])
    assert numLevels >= 1
    if size is None:
        size = minSize + random.choice([randint(3), randint(6)])

    alignment = []
    startTime = startTimeInit
    for _ in range(size):
        label = gen_label(minSize=labelMinSize, alphabet=labelAlphabet)
        if numLevels == 1:
            dur = minDur + random.choice([randint(3), randint(10)])

            endTime = startTime + dur
            subAlignment = None
        else:
            subAlignment = gen_alignment(
                startTimeInit=startTime,
                numLevels=(numLevels - 1),
                size=None,
                minSize=minSizeSub,
                minSizeSub=minSizeSub,
                minDur=minDur,
                labelMinSize=labelMinSize,
                labelAlphabet=labelAlphabet,
            )
            if not subAlignment:
                endTime = startTime
            else:
                endTime = subAlignment[-1][1]

        alignment.append((startTime, endTime, label, subAlignment))
        startTime = endTime

    return alignment

def gen_framePeriod():
    """Returns a frame period.

    Specifically, returns a frame period such that writing-then-reading
    alignments should be lossless.
    """
    while True:
        framePeriod = abs(randn()) * 3e-7
        if framePeriod >= 1e-7:
            break

    return framePeriod

class SimpleLabelMap(object):
    def __init__(self, ident):
        self.ident = ident

    def __call__(self, label):
        return self.ident, label

def gen_labelMap():
    ident = randint(1000)
    return SimpleLabelMap(ident)

class AlignmentTest(unittest.TestCase):
    def test_SimpleAlignmentIo_writeLines_empty(self):
        framePeriod = gen_framePeriod()
        alignmentIo = alio.SimpleAlignmentIo(framePeriod)
        self.assertEqual(alignmentIo.writeLines([]), [])

    def test_SimpleAlignmentIo_writeLines_single_line(self, its=50):
        for it in range(its):
            framePeriod = gen_framePeriod()
            alignmentIo = alio.SimpleAlignmentIo(framePeriod)
            startTime = randint(-10, 11)
            dur = randint(10)
            endTime = startTime + dur
            label = gen_label()
            alignment = [(startTime, endTime, label, None)]

            alignmentLines = alignmentIo.writeLines(alignment)

            startTicks = int(round(startTime * framePeriod * 1e7))
            endTicks = int(round(endTime * framePeriod * 1e7))
            alignmentLinesGood = ['%s %s %s' % (startTicks, endTicks, label)]
            self.assertEqual(alignmentLines, alignmentLinesGood)

    def test_SimpleAlignmentIo_writeLines_concat(self, its=50):
        for it in range(its):
            framePeriod = gen_framePeriod()
            alignmentIo = alio.SimpleAlignmentIo(framePeriod)
            alignment = gen_alignment()
            breakIndex = randint(len(alignment) + 1)
            alignment0 = alignment[:breakIndex]
            alignment1 = alignment[breakIndex:]

            alignmentLines0 = alignmentIo.writeLines(alignment0)
            alignmentLines1 = alignmentIo.writeLines(alignment1)
            alignmentLines = alignmentIo.writeLines(alignment)
            self.assertEqual(alignmentLines0 + alignmentLines1, alignmentLines)

    def test_SimpleAlignmentIo_readLines(self, its=50):
        for it in range(its):
            alignment = gen_alignment()
            framePeriod = gen_framePeriod()
            alignmentIo = alio.SimpleAlignmentIo(framePeriod)

            alignmentLines = alignmentIo.writeLines(alignment)
            alignmentAgain = alignmentIo.readLines(alignmentLines)
            self.assertEqual(alignmentAgain, alignment)

    def test_flatten_1_level(self, its=50):
        for it in range(its):
            alignment = gen_alignment()

            flatAlignment = alio.flatten(alignment)

            flatAlignmentGood = [
                (startTime, endTime, (label,), None)
                for startTime, endTime, label, subAlignment in alignment
            ]
            self.assertEqual(flatAlignment, flatAlignmentGood)

    def test_flatten_nested(self, its=50):
        for it in range(its):
            numLevels = randint(2, 4)
            alignment = gen_alignment(numLevels=numLevels)

            flatAlignment = alio.flatten(alignment)

            flatAlignmentNested = alio.flatten([
                (startTime, endTime, label, alio.flatten(subAlignment))
                for startTime, endTime, label, subAlignment in alignment
            ])

            def getLabelTuple(labelTupleTuple):
                if len(labelTupleTuple) == 1:
                    labelTuple = labelTupleTuple[0]
                else:
                    labelRest, labelOuter = labelTupleTuple
                    labelTuple = labelRest + (labelOuter,)
                return labelTuple

            flatAlignmentAgain = [
                (
                    startTime, endTime, getLabelTuple(labelTupleTuple),
                    subAlignment
                )
                for startTime, endTime, labelTupleTuple, subAlignment in (
                    flatAlignmentNested
                )
            ]
            self.assertEqual(flatAlignmentAgain, flatAlignment)

    def test_unflatten(self, its=50):
        for it in range(its):
            numLevels = randint(1, 4)
            alignment = gen_alignment(numLevels=numLevels)

            flatAlignment = alio.flatten(alignment)
            alignmentAgain = alio.unflatten(flatAlignment)
            self.assertEqual(alignmentAgain, alignment)

    def test_AlignmentIo_writeLines_1_level(self, its=50):
        for it in range(its):
            framePeriod = gen_framePeriod()
            alignment = gen_alignment()
            simpleAlignmentIo = alio.SimpleAlignmentIo(framePeriod)
            alignmentIo = alio.AlignmentIo(framePeriod)

            self.assertEqual(
                simpleAlignmentIo.writeLines(alignment),
                alignmentIo.writeLines(alignment)
            )

    def test_AlignmentIo_writeLines_nested(self, its=50):
        for it in range(its):
            numLevels = randint(2, 4)
            alignment = gen_alignment(numLevels=numLevels)
            framePeriod = gen_framePeriod()
            alignmentIo = alio.AlignmentIo(framePeriod)

            alignmentLines = alignmentIo.writeLines(alignment)

            alignmentLinesAgain = []
            for startTime, endTime, label, subAlignment in alignment:
                subAlignmentLines = alignmentIo.writeLines(subAlignment)
                alignmentLinesAgain.append(subAlignmentLines[0] + ' ' + label)
                alignmentLinesAgain.extend(subAlignmentLines[1:])

            self.assertEqual(alignmentLinesAgain, alignmentLines)

    def test_AlignmentIo_readLines(self, its=50):
        for it in range(its):
            numLevels = randint(1, 4)
            alignment = gen_alignment(numLevels=numLevels)
            framePeriod = gen_framePeriod()
            alignmentIo = alio.AlignmentIo(framePeriod)

            alignmentLines = alignmentIo.writeLines(alignment)
            alignmentAgain = alignmentIo.readLines(alignmentLines)
            self.assertEqual(alignmentAgain, alignment)

    def test_mapAlignmentLabels_1_level(self, its=50):
        for it in range(its):
            alignment = gen_alignment()
            labelMap = gen_labelMap()

            alignmentNew = alio.mapAlignmentLabels(alignment, [labelMap])

            alignmentNewGood = [
                (startTime, endTime, labelMap(label), None)
                for startTime, endTime, label, subAlignment in alignment
            ]
            self.assertEqual(alignmentNew, alignmentNewGood)

    def test_mapAlignmentLabels_nested(self, its=50):
        for it in range(its):
            numLevels = randint(2, 4)
            alignment = gen_alignment(numLevels=numLevels)
            labelMaps = [ gen_labelMap() for _ in range(numLevels) ]

            alignmentNew = alio.mapAlignmentLabels(alignment, labelMaps)

            alignmentNewAgain = [
                (
                    startTime, endTime, labelMaps[0](label),
                    alio.mapAlignmentLabels(subAlignment, labelMaps[1:])
                )
                for startTime, endTime, label, subAlignment in alignment
            ]

            self.assertEqual(alignmentNewAgain, alignmentNew)

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/python
"""A setuptools-based script for distributing and installing htk_io."""

# Copyright 2014, 2015 Matt Shannon

# This file is part of htk_io.
# See `License` for details of license and warranty.


import os
from setuptools import setup

with open('README.rst') as readmeFile:
    long_description = readmeFile.read()

requires = [ line.rstrip('\n') for line in open('requirements.txt') ]

setup(
    name='htk_io',
    version='0.5',
    description='Read and write HTK and HTS files from python.',
    url='http://github.com/MattShannon/htk_io',
    author='Matt Shannon',
    author_email='matt.shannon@cantab.net',
    license='3-clause BSD (see License file)',
    packages=['htk_io'],
    install_requires=requires,
    scripts=[
        'bin/htk_io_get_label_map_leaf_macro_id_to_leaf_index.py',
        'bin/htk_io_map_alignment_files_label_sublabel_to_leaf_macro_id.py',
        'bin/htk_io_map_alignment_files.py',
    ],
    long_description=long_description,
)



"""Functions for reading HMGenS-style "pdf" files.

When the HMGenS tool in HTS is run with the '-p' flag, it produces a
probability density function (pdf) file for each utterance and stream.
An HMGenS pdf file specifies Gaussian parameters for each frame, window and
parameter index.
The Gaussian parameters are the b-value (also known as mean-times-precision)
and the precision (also known as inverse variance).

Note that the above interpretation of the values in a HMGenS pdf file as the
b-value and precision of a Gaussian distribution is only valid for the
"standard" model used while training a statistical parametric speech synthesis
system.
This model is not actually a proper probabilistic model of speech parameter
sequences since it models the speech parameter trajectory for each window
separately.
For proper probabilistic models such as the trajectory HMM, the values in an
HMGenS pdf file can be interpreted in a related but different way.
"""

# Copyright 2014, 2015 Matt Shannon

# This file is part of htk_io.
# See `License` for details of license and warranty.

import numpy as np

# FIXME : add tests

# FIXME : change to a HMGenSPdfIo class
def readHMGenSPdf(pdfFile, paramOrder, numWindows):
    """Reads an HMGenS probability density (pdf) file into a numpy array.

    Returns two arrays, each of shape (numTimes, numWindows, paramOrder), the
    first specifying the b-value and the second specifying the precision.
    """
    pdfAll = np.reshape(
        np.fromfile(pdfFile, dtype=np.float32),
        (-1, 2, numWindows, paramOrder)
    ).astype(np.float)
    bWinAll = pdfAll[:, 0]
    tauWinAll = pdfAll[:, 1]
    return bWinAll, tauWinAll



"""Functions for reading and writing HTK / HTS decision tree files."""

# Copyright 2014, 2015 Matt Shannon

# This file is part of htk_io.
# See `License` for details of license and warranty.

from collections import deque

from htk_io.misc import stripQuotes, addQuotes, verifiedRead
import htk_io.ques as qio

class Leaf(object):
    def __init__(self, macroId):
        self.macroId = macroId

    def __str__(self):
        return addQuotes(self.macroId)

    def __repr__(self):
        return 'Leaf(%r)' % self.macroId

class Tree(object):
    def __init__(self, splitInfos, rootNode):
        self.getQuesId = dict()
        self.getChildren = dict()
        self.splitIdsInOrigOrder = [
            splitId
            for splitId, quesId, leftChild, rightChild in splitInfos
        ]
        for splitId, quesId, leftChild, rightChild in splitInfos:
            self.getQuesId[splitId] = quesId
            self.getChildren[splitId] = [leftChild, rightChild]

        self.rootNode = rootNode
        assert (isinstance(self.rootNode, Leaf) or
                self.rootNode in self.getChildren)

        self.nodes = [ node for node, depth in self.breadthFirst() ]
        self.splitIds = [ node
                          for node in self.nodes
                          if not isinstance(node, Leaf) ]
        assert len(self.splitIds) <= len(splitInfos)
        if len(self.splitIds) < len(splitInfos):
            raise RuntimeError('there appear to be unreachable nodes')
        self.leaves = [ node
                        for node in self.nodes
                        if isinstance(node, Leaf) ]
        assert len(self.leaves) == len(self.splitIds) + 1

    def breadthFirst(self, nodeStart = None):
        """Returns an iterator over nodes in the tree in breadth-first order.

        Each element of the returned iterator is a pair of a node and its
        depth.
        """
        if nodeStart is None:
            nodeStart = self.rootNode
        agenda = deque([(nodeStart, 0)])
        while agenda:
            node, depth = agenda.popleft()
            yield node, depth
            if not isinstance(node, Leaf):
                for childNode in self.getChildren[node]:
                    agenda.append((childNode, depth + 1))

    def breadthFirstWithCIL(self, nodeStart = None):
        """Returns an iterator over nodes in the tree in breadth-first order.

        Each element of the returned iterator is a pair of a node and the
        sequence of child indices corresponding to the path taken in the tree
        to get to that node.
        """
        if nodeStart is None:
            nodeStart = self.rootNode
        agenda = deque([(nodeStart, [])])
        while agenda:
            node, childIndexList = agenda.popleft()
            yield node, childIndexList
            if not isinstance(node, Leaf):
                for childIndex, childNode in enumerate(self.getChildren[node]):
                    childIndexListNew = list(childIndexList)
                    childIndexListNew.append(childIndex)
                    agenda.append((childNode, childIndexListNew))

# (FIXME : "navigable" is a horrible term for this and concept seems bad too)
class NavBinaryTree(object):
    """A "navigable" binary decision tree.

    "Navigable" here means that the tree knows the "meaning" of its questions,
    so we can navigate the tree for a given label.

    Each internal node in the tree should have two children, with the first
    child corresponding to a "no" answer to that node's question and the second
    child corresponding to a "yes" answer.
    """
    def __init__(self, quesReDict, tree):
        self.quesReDict = quesReDict
        self.tree = tree

        for node in self.tree.splitIds:
            assert len(self.tree.getChildren[node]) == 2

    def getLeaf(self, label):
        """Returns the leaf associated with a label."""
        node = self.tree.rootNode
        while not isinstance(node, Leaf):
            quesId = self.tree.getQuesId[node]
            children = self.tree.getChildren[node]
            quesRe = self.quesReDict[quesId]
            node = children[1] if quesRe.match(label) else children[0]

        return node

def readTreeFileLines(treeFileLines):
    questions, treeFileLines = qio.parseQuestionLines(
        treeFileLines, isTreeFile=True
    )

    streamSpecedTrees = []

    currStreamSpec = None
    currSplitInfos = None

    for line in treeFileLines:
        line = line.strip()
        fields = line.split()
        if len(fields) > 0:
            assert fields[0] != 'QS'

            if currStreamSpec is None:
                assert len(fields) == 1
                assert currSplitInfos is None
                currStreamSpec = fields[0]
            elif fields[0] == '{':
                assert len(fields) == 1
                assert currStreamSpec is not None
                assert currSplitInfos is None
                currSplitInfos = []
            elif fields[0] == '}':
                assert len(fields) == 1
                assert currStreamSpec is not None
                assert currSplitInfos is not None
                streamSpecedTrees.append((
                    currStreamSpec,
                    Tree(currSplitInfos, rootNode = 0)
                ))
                currStreamSpec = None
                currSplitInfos = None
            elif len(fields) == 1 and fields[0][0] == '"':
                # degenerate tree with just a root leaf
                assert currStreamSpec is not None
                assert currSplitInfos is None
                currSplitInfos = []
                macroId = stripQuotes(fields[0])
                streamSpecedTrees.append((
                    currStreamSpec,
                    Tree(currSplitInfos, rootNode = Leaf(macroId))
                ))
                currStreamSpec = None
                currSplitInfos = None
            else:
                assert len(fields) == 4
                assert currStreamSpec is not None
                assert currSplitInfos is not None
                splitId = int(fields[0])
                quesId = fields[1]
                if fields[2][0] == '"':
                    macroId = stripQuotes(fields[2])
                    leftChild = Leaf(macroId)
                else:
                    leftChild = int(fields[2])
                if fields[3][0] == '"':
                    macroId = stripQuotes(fields[3])
                    rightChild = Leaf(macroId)
                else:
                    rightChild = int(fields[3])
                currSplitInfos.append((
                    splitId, quesId,
                    leftChild, rightChild
                ))

    quesIdsSet = set([ quesId for quesId, _ in questions ])
    assert len(quesIdsSet) == len(questions)
    for streamSpec, tree in streamSpecedTrees:
        for splitId in tree.splitIds:
            quesId = tree.getQuesId[splitId]
            assert quesId in quesIdsSet

    return questions, streamSpecedTrees

def writeTreeFileLines(questions, streamSpecedTrees):
    outLines = list(qio.writeQuestionLines(questions, isTreeFile=True))
    for streamSpec, tree in streamSpecedTrees:
        outLines.append('')
        outLines.append(' %s' % streamSpec)
        if isinstance(tree.rootNode, Leaf):
            # degenerate tree with just a root leaf
            outLines.append(' %s' % tree.rootNode)
        else:
            outLines.append('{')
            for splitId in tree.splitIdsInOrigOrder:
                quesId = tree.getQuesId[splitId]
                leftChild, rightChild = tree.getChildren[splitId]
                outLines.append(
                    ' %s %s %s %s' %
                    (splitId, quesId, leftChild, rightChild)
                )
            outLines.append('}')
    outLines.append('')
    return outLines

def readTreeFile(treeFile):
    """Reads a decision tree file."""
    lines = [ line.rstrip('\n') for line in open(treeFile, 'U') ]
    questions, streamSpecedTrees = readTreeFileLines(lines)
    return questions, streamSpecedTrees

def readTreeFileVerifying(treeFile):
    """Reads a decision tree file and verifies that it was read correctly.

    Verifies that reconstructing the decision tree file from the parsed output
    reproduces the original decision tree file up to certain whitespace
    substitutions.
    """
    lines = [ line.rstrip('\n') for line in open(treeFile, 'U') ]
    questions, streamSpecedTrees = verifiedRead(
        readTreeFileLines, writeTreeFileLines, lines, unpack=True
    )
    return questions, streamSpecedTrees

def writeTreeFile(questions, streamSpecedTrees, treeFile):
    """Writes a decision tree file."""
    lines = writeTreeFileLines(questions, streamSpecedTrees)
    with open(treeFile, 'w') as f:
        for line in lines:
            f.write(line)
            f.write('\n')
